
from allensdk.api.queries.rma_api import RmaApi

import pandas as pd
import numpy as np

from . import Utils
import copy 

Z_SCORE = 'z-score'
EXPR_LVL = 'expression_level'
REGION_ASSIGNMENT = 'mapped region'

VALUE_COLUMNS = [EXPR_LVL,Z_SCORE] 

DATAFRAME_CACHE = "cache\\data-frames\\"

Genes = ["Gabra1", "Gabra2", "Gabra4", "Gabra5", "Gabrb1", "Gabrb2", "Gabrb3", "Gabrd", "Gabrg2", "Gabrg3"]

# ! here, some magic happens. these lists define - amongst the available options in dropdowns, their labels, and defaults - also some interactions with the charts.
# ! the type defines the named parameter that is provided to the chart-functions. 
GENE_LIST = Utils.DropDownListConfiguration(label='Gene', type='gene', 
  data=Genes,
  default='Gabra4', defaultLeft='Gabra4', defaultRight='Gabrb3')

# for co-expressions, we need additional gene-selection. this ensure that the type is correctly bound to the expected parameters
GENE1_LIST = Utils.DropDownListConfiguration(label='Gene', type='gene1', 
  data=Genes,
  default='Gabra4', defaultLeft='Gabra4', defaultRight='Gabra4')

GENE2_LIST = Utils.DropDownListConfiguration(label='vs', type='gene2', 
  data=Genes,
  default='Gabra4', defaultLeft='Gabra5', defaultRight='Gabra5')

AGGREGATION_FUNCTIONS = Utils.DropDownListConfiguration(label='Aggregation function', type='aggregation_function', 
  data=['min', 'max', 'mean', 'var'], 
  default='mean', defaultLeft='mean',defaultRight='var')

SPECIES = Utils.DropDownListConfiguration(label=None, type='species', 
  data=['human', 'mouse - sagittal', 'mouse - coronal'], 
  default='human', defaultLeft= 'human', defaultRight= 'mouse - sagittal')

STRUCTURE_LEVELS = Utils.DropDownListConfiguration(label='Level', type='structure_level', 
  data=[l for l in range(0,10)], 
  default=2, defaultLeft= 2, defaultRight= 2)

# this is used to define how to aggregate already aggregated values. e.g., aggregating the (expression-level, max)-column uses np.max.
AGGREGATION_AGGREGATES = { 'min': np.min, 'max': np.max, 'mean': np.mean, 'var': np.mean }

class AllenSdkHelper:
  def __init__(self):
    self.rma = RmaApi() 
    
    # TODO: only load this, if the file does not exist!
    self.PlaneOfSections = self.rma.json_msg_query(
            url="http://api.brain-map.org/api/v2/data/query.json?criteria=model::PlaneOfSection,rma::options[num_rows$eqall]"
        )

    # path=Utils.makedir(f'cache\\models') + '\\PlaneOfSection.json',

  def getPlaneOfSections(self):
    return self.PlaneOfSections

allenSdkHelper = AllenSdkHelper()

PlaneOfSections = {x['id']: x['name'] for x in allenSdkHelper.getPlaneOfSections()} 

__opposing = { 'Human': 'Mouse', 'Mouse': 'Human' }

__regionAssignmentsRaw = pd.read_csv('region-assignments.csv', header=0)
__regionAssignments = { species: __regionAssignmentsRaw.apply(lambda x: 
    { (x[species].split(';')[0], x[species].split(';')[1]) :
     { 'assignment': (x[__opposing[species]].split(';')[0], x[__opposing[species]].split(';')[1]) ,
     'name': x['Name'] }
    } ,axis=1)
     for species in ['Human', 'Mouse'] }

# https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_dict.html
RegionAssignments = Utils.simple( { 
  'asList': __regionAssignments,
  'asDict': { 
    'Human': Utils.combine_dicts(__regionAssignments['Human'].to_list()),
    'Mouse': Utils.combine_dicts(__regionAssignments['Mouse'].to_list()) }
})



from . import Visualisation
from . import Constants

def setAvailableGenes(str):
  Constants.Genes = str

def start():
  webApp = Visualisation.WebInterface(__name__) 

  webApp.run_server(debug=False)


from allensdk.api.queries.rma_api import RmaApi

from types import SimpleNamespace

import numpy as np
import pandas as pd
import glob

import concurrent
import concurrent.futures

from . import Utils
from . import Constants

from . import StructureMap

class HumanMicroarrayData:
  VALUE_COLUMNS = [Constants.EXPR_LVL, Constants.Z_SCORE] 

  currentGets = {}
  
  def __init__(self, geneAcronym):
      self.geneAcronym = geneAcronym
      self.cache_path = Constants.DATAFRAME_CACHE + f'human\\{geneAcronym}\\'
  
  # ok, now we got n probes with m expression-levels & z-scores
  # we also got m samples that describe which donor and which structure each expression-level stems from
  # we have to be aware that the expression-levels are retrieved from a probe, which represents a plane through the brain.
  # so if the plane of the probe is not cutting through a specific brain-region, then there are null-values present for the expression-level.
  # details: http://help.brain-map.org/display/humanbrain/API

  def transformExpressionData(self, expressionData):
    
    # this class allows us to add attributes to an object
    # https://docs.python.org/3/library/types.html#types.SimpleNamespace
    combined = SimpleNamespace()

    setattr(combined, 'samples', []) 
    setattr(combined, Constants.EXPR_LVL, [])
    setattr(combined, 'z_score', [])
    
    samples = expressionData["samples"] # we hereby prevent a dict-lookup for each probe, because its always the same data used over and over again

    for probe in expressionData["probes"]:
      # https://stackoverflow.com/questions/30522724/take-multiple-lists-into-dataframe
      # we need to map each probe to the sample-annotations (see MicroarrayData_Readme.txt, provided by the Allen Institue).
      # so, we basically repeat the samples for each probe:
      combined.samples += samples

      # these are provided in the same strucutural manner
      combined.expression_level += probe[Constants.EXPR_LVL]
      combined.z_score += probe[Constants.Z_SCORE] 

      # the z-scores provided here come with some side-notes, according to http://help.brain-map.org/display/humanbrain/API 
      # see: https://community.brain-map.org/t/z-score-for-human-microarray-and-mouse-ish-data/912/3
      # and: https://community.brain-map.org/t/reproducing-r-score-correlations-in-allen-human-brain-atlas/910/4
      # key take-away: z-scores are calculated on behalf of the expression-levels per donor. 
      # for mice, we only have 1 donor per experiment, so we are fine by calculating z-scores for mice ourselves.
      # for humans, we rely on the values provided by the Allen Institute

    # https://stackoverflow.com/questions/29325458/dictionary-column-in-pandas-dataframe
    data = pd.DataFrame({Constants.EXPR_LVL: combined.expression_level, Constants.Z_SCORE: combined.z_score},
                                dtype=np.float32) # setting this type is important for later aggregation. else, pandas throws an error for mean & var

    # the sample's metadata is stored as dictionary-entries. we unpack them using this function, in order to transform them into columns
    def unpack_dict_list(dict_list, attribute, prefix):
      # read it like this: each entry in a list of dictionaries (e.g. combined.samples) is mapped to isolate a specific attribute (e.g. 'structure').
      # then, create a dataframe from this list. lastly, add the provided prefix in order to prevent naming conflicts.
      return pd.DataFrame.from_dict(map(lambda x: x[attribute], dict_list)).add_prefix(prefix) 

    # attributes with their respective prefix to prevent ambiguous column-names
    attributes = [("donor", ""), ("sample", "sample_"), ("structure", "structure_")] 

    # note that here, the * is the splat-operator. it is used to unpack the array produced by the list comprehension,
    # in order to provide pd.concat with a plain list of dataframes to concat.
    data = pd.concat([*[unpack_dict_list(combined.samples, attr[0], attr[1]) for attr in attributes], data], axis=1)

    # dropna is super slow, so we use this approach instead:
    data = data[data[Constants.EXPR_LVL].notnull() & data[Constants.Z_SCORE].notnull()]

    return data 
  
  def get(self, from_cache, aggregations):
  
    if from_cache:
      return self.getAsync(from_cache, aggregations)

    if self.geneAcronym in HumanMicroarrayData.currentGets:
      print(f'Waiting for initial request of human gene {self.geneAcronym} to complete...')
      done, not_done = concurrent.futures.wait([HumanMicroarrayData.currentGets[self.geneAcronym]], 
       return_when=concurrent.futures.FIRST_COMPLETED) # this wants an array... ok
      
      for fut in done:
        print(fut, 'exception:', fut.exception())
        return fut.result() 

    else: 
      with concurrent.futures.ThreadPoolExecutor() as executor:
        HumanMicroarrayData.currentGets[self.geneAcronym] = executor.submit(self.getAsync, from_cache, aggregations)
        return HumanMicroarrayData.currentGets[self.geneAcronym].result()

  #@Utils.profile(sort_by='cumulative', lines_to_print=10, strip_dirs=True)
 
  def getAsync(self, from_cache, aggregations): # load data once with use_cache = True, then change it to False to read it from disk instead of fetching it from the api
    #print('HumanMicroarrayData.get() start')
    if not from_cache:
      # we use the RmaApi to query specific information, such as the section data sets of a specific gene
      # for docs, see: https://alleninstitute.github.io/AllenSDK/allensdk.api.queries.rma_api.html
      rma = RmaApi() 

      # ok, so we don't need to do multiple requests to forward data from a model to a service, but simply use the pipe-concept:
      # http://help.brain-map.org/display/api/Service+Pipelines
      # e.g. this finds all probes for gabra4 and then queries the microarray-expression data for these probes. note that variables generated by a pipe are referenced by $variableName

      # check out this playground: http://api.brain-map.org/examples/rma_builder/index.html
      # we only use the product 'Human Microarray', which has the id 2 (id$eq2).
      query = ("http://api.brain-map.org/api/v2/data/query.json?criteria="
              f"model::Probe,rma::criteria,gene[acronym$il{self.geneAcronym}],products[id$eq2],rma::options[num_rows$eqall],"
              "pipe::list[probes$eq'id'],"
              "service::human_microarray_expression[probes$eq$probes]")
      
      data = rma.json_msg_query(url=query)

      structure_map  = StructureMap.StructureMap(reference_space_key = 'annotation/ccf_2017', resolution = 25).get(structure_graph_id=10) # , annotation, meta 
      
      data = self.transformExpressionData(data)

      # https://stackoverflow.com/questions/19125091/pandas-merge-how-to-avoid-duplicating-columns
      # to avoid automatic renaming the duplicate columns by removing any duplicate-column
      # note that our merge-condition is index vs structure_id. because structure_id is the index of structure_map, 
      # it is not identified as a duplicate column.
      data = data[data.columns.difference(structure_map.columns)]

      ret = Utils.merge_with_structure(data, structure_map, HumanMicroarrayData.VALUE_COLUMNS, aggregations)
      
      Utils.save(ret, self.cache_path, 'cache.pkl')
      
      return { 'human': ret, } 

    else:
      if not glob.glob(self.cache_path):
        Utils.log.warning(f"No cached dataframe found. Check whether you have access to file '{self.cache_path}' and whether it exists. Obtaining data without caching now...")
        return self.get(False, aggregations)

      #print('HumanMicroarrayData.get() done')
      return { 'human': Utils.load(self.cache_path + 'cache.pkl') }



# this file is required in order to be able to debug the app, e.g. under vs code. 
# make sure to define "module": "src.geneEcomparison" in launch.json - else, the package wont be available
# the app is only started when run as a script
# https://docs.python.org/3/library/__main__.html
if __name__ == "__main__":
    # execute only if run as a script
  from . import App
  App.start()



# used for data-visualisation
import plotly.graph_objects as go
import plotly.express as px

# library for providing an interactive web-interface (react-based)
import dash
from dash import Dash, no_update
import dash_bootstrap_components as dbc # used for css
import dash_core_components as dcc 
import dash_html_components as html # basic html-components
from dash.dependencies import Input, Output, MATCH, ALL # used for callback-functions
from dash_extensions import Download # provide general download-capabilities
from dash_extensions.snippets import send_data_frame # helps us provide downloadable data-frames

import dash_table # the component for showing data in a grid. also see: https://community.plotly.com/t/loading-pandas-dataframe-into-data-table-through-a-callback/19354/14

import webbrowser
from threading import Timer
import traceback # used for printing the stacktrace of callback-exceptions

from .HumanMicroarrayData import HumanMicroarrayData
from .MouseISHData import MouseISHData

from . import Comparison
from .Constants import AGGREGATION_AGGREGATES, AGGREGATION_FUNCTIONS, GENE1_LIST, GENE2_LIST, GENE_LIST, SPECIES, STRUCTURE_LEVELS, Z_SCORE, EXPR_LVL
from . import Utils

import sys
import json


# we use some icons from the font-awesome library:
FONT_AWESOME = "https://use.fontawesome.com/releases/v5.7.2/css/all.css"

# ! adjust lists declared in Constants.py (such as GENE_LIST) to change the dropdown-list's options.

# ? For some good tutorials on Dash (https://dash.plotly.com/), see:
# ? https://www.youtube.com/watch?v=Ldp3RmUxtOQ
# ? https://www.youtube.com/watch?v=hSPmj7mK6ng

class WebInterface:
  def __init__(self, name, port = 5000):

    # for layout & css, see https://dash.plotly.com/layout
    self.app = dash.Dash(__name__, external_stylesheets=[dbc.themes.LUX, FONT_AWESOME])
    self.port = port
   
    self.app.title = "Gene-expression comparison"
    self.header = [html.Span('Z-scores', id="z_score", style={"textDecoration": "underline", "cursor": "pointer"}), 
      # we need to use unicode, because html cannot be rendered without circumventing react's html-sanitizer (which is a security-feature)
      # formatting with line-breaks: https://community.plotly.com/t/dash-bootstrap-components-tooltip-multiple-lines/42285/2
      dbc.Tooltip(dcc.Markdown("= (\u0078 - \u03bc) / \u03C3\n\nMeasures how many standard-deviations the given value is afar from the population's mean."), target="z_score", style={"font-size": "16px"}), 
      ' of gene-expression by species & region']
    self.credits ="by Christoph Hüsson"
    self.description = [
      "All data obtained from Allen Brain Institute: Human microarray-data vs rodent in-situ hybridization. Note that 'mouse - sagittal' only provides data for left hemisphere. However, for some genes, coronal data is not available for mice.",
      html.Br(), 
      "Double-click on a legend-entry to isolate associated data-points, e.g. to only view co-expressions of a specific region."]
    self.downloads = {}

    self.loadingColor = "#88888888"
    VIEWS = Utils.simple({
      'coexpressions': Utils.simple({ 'name': 'coexpressions', 'fn': self.coexpressions}), 
      'stackedBarsBySpecies': Utils.simple({'name': 'stackedBarsBySpecies', 'fn': self.stackedBarsBySpecies}),
      'allSpecies': Utils.simple({ 'name': 'allSpecies', 'fn': self.heatmapByRegion}),
      'gridView': Utils.simple({ 'name': 'gridView', 'fn': self.getTableData})
    })

    # https://dash-bootstrap-components.opensource.faculty.ai/docs/components/layout/
    self.app.layout = html.Div([
        html.H1(self.header, className="pl-4 pt-2"),
        html.Span(self.credits, className="pl-4 text-muted", style={ 'position': 'absolute', 'top': '0.5rem', 'right': '0.5rem'}),
        html.P(self.description, className="pl-4"),
        # TODO: explain z-score with a tooltip: https://dash-bootstrap-components.opensource.faculty.ai/docs/components/tooltip/ 
        dbc.Tabs([
          dbc.Tab(self.sideBySideView(VIEWS.allSpecies.name, [AGGREGATION_FUNCTIONS], [GENE_LIST]),  
            label="Expression-heatmaps"),
          dbc.Tab(self.sideBySideView(VIEWS.coexpressions.name, [AGGREGATION_FUNCTIONS], [SPECIES, GENE1_LIST, GENE2_LIST, STRUCTURE_LEVELS]),
           label="Co-expressions"),
          dbc.Tab(self.sideBySideView(VIEWS.stackedBarsBySpecies.name, [AGGREGATION_FUNCTIONS], [SPECIES, GENE_LIST, STRUCTURE_LEVELS]), 
           label="Stacked bars"),
          dbc.Tab(self.gridView(VIEWS.gridView.name, [SPECIES, GENE_LIST, STRUCTURE_LEVELS]), label="Data-grid")
        ]),
    ])

    @self.app.server.route('/shutdown', methods=['GET'])
    def shutdown():
      sys.exit();
      return 'Server shutting down...'

    # from: https://community.plotly.com/t/allowing-users-to-download-csv-on-click/5550/42
    @self.app.callback(Output({'type': 'download', 'view': MATCH}, "data"), Input({'type': 'download-button', 'view': MATCH}, "n_clicks"))
    def downloadCallback(n_clicks):
      if not n_clicks is None:
        dl = self.downloads['grid']

        return send_data_frame(dl['data'].to_excel, dl['filename'] + '.xlsx', index=True)
      else:
        return no_update

    # we need to update left/right side https://dash.plotly.com/pattern-matching-callbacks
    # see https://dash-bootstrap-components.opensource.faculty.ai/docs/components/input/
    @self.app.callback( 
      # MATCH is a bit tricky. in this case, it basically says that all inputs and outputs need to belong to the same view
      # for a doc on available output-properties (figure, hidden, etc.), check out: https://dash.plotly.com/dash-html-components/output
      [Output({'type': 'graph', 'side': 'left', 'view': MATCH}, "figure"), 
      # graphs do NOT have a 'hidden'-property. setting their style in the callback would override previously set style-settings, such as the graph's dimensions.
      # instead, we use a wrapper-div, which provides us with a hidden property.
       Output({'type': 'graph-container', 'side': 'left', 'view': MATCH}, "hidden"), 
       Output({'type': 'graph', 'side': 'right', 'view': MATCH}, "figure"),
       Output({'type': 'graph-container', 'side': 'right', 'view': MATCH}, "hidden"),
       Output({'type': 'alert', 'side': 'left', 'view': MATCH}, "children"),
       Output({'type': 'alert', 'side': 'left', 'view': MATCH}, "is_open"),
       Output({'type': 'alert', 'side': 'right', 'view': MATCH}, "children"),
       Output({'type': 'alert', 'side': 'right', 'view': MATCH}, "is_open")], 
       # we need to separate the side-specific graphs from the inputs. we use input: True to filter out non-input-controls
      [Input({'type': ALL, 'input': True, 'view': MATCH}, "value"), 
       Input({'type': ALL, 'input': True, 'side': 'left', 'view': MATCH}, "value"), 
       Input({'type': ALL, 'input': True, 'side': 'right', 'view': MATCH}, "value")])
    def sideBySideViewCallback(common, left, right): # these parameters are actually not required / usable, because the real inputs are provided dynamically
      
      # we don't want to update a side if only the opposing side's gene has been changed
      # https://dash.plotly.com/advanced-callbacks
      ctx = dash.callback_context

      # https://stackoverflow.com/questions/4547274/convert-a-python-dict-to-a-string-and-back
      # the prop_id is a stringified dictionary (ugh...) so we need to json.loads it. but it is provided in the format 'stringified_dict.value', so we split it...
      # ! we cant use eval here, because Dash lowercases the property-values, which changes True to true. eval does not recognize this as a keyword and crashes
      input_props = json.loads(ctx.triggered[0]['prop_id'].split('.')[0]) if ctx.triggered else {}

      common = self.callbackInputsToKwags(ctx.inputs_list[0])
      left = self.callbackInputsToKwags(ctx.inputs_list[1])
      right = self.callbackInputsToKwags(ctx.inputs_list[2])

      # handle the initial load, which is provided with triggered = False. changing non-side related inputs, however, forces a reload of both sides
      if not ctx.triggered or 'side' not in input_props: 
        left_unchanged = False
        right_unchanged = False
      else:
        # check which input triggered this callback. there can only be one input at a time, so we only check the first element:
        left_unchanged = (input_props['side'] != "left")
        right_unchanged = (input_props['side'] != "right")

      # the function used for creating the chart-figure is defined as a pointer in the views-object:
      fn = getattr(VIEWS, ctx.outputs_list[0]['id']['view']).fn

      retLeft = None
      retRight = None 

      errLeft = None
      errRight = None

      if not left_unchanged:
        try:
          retLeft = fn(**common, **left, side='left')
        except Exception as e:
          # https://stackoverflow.com/questions/4308182/getting-the-exception-value-in-python
          traceback.print_exc()
          errLeft = str(e)

      if not right_unchanged:
        try:
          retRight = fn(**common, **right, side='right')
        except Exception as e:
          traceback.print_exc()
          errRight = str(e)

      # https://community.plotly.com/t/i-want-to-create-a-conditional-callback-in-dash-is-it-possible/23418/2
      # we are able to prevent an update of one side. however, the loading-indicator will still be shown. 
      # track this bug here: https://github.com/plotly/dash/issues/1120

      return (
        {} if errLeft else no_update if left_unchanged else retLeft, 
        not not errLeft,
        {} if errRight else no_update if right_unchanged else retRight,
        not not errRight,
        no_update if errLeft is None else errLeft, # alert-text
        not (errLeft is None), # alert is_open
        no_update if errRight is None else errRight,
        not (errRight is None)
        )

        
    @self.app.callback( 
      # MATCH is a bit tricky. in this case, it basically says that all inputs and outputs need to belong to the same view
      Output({'type': 'grid', 'view': MATCH}, "children"), 
        # we need to separate the side-specific graphs from the inputs. we use input: True to filter out non-input-controls
      Input({'type': ALL, 'input': True, 'view': MATCH}, "value"))
    def gridViewCallback(common): # these parameters are actually not required / usable, because the real inputs are provided dynamically
      
      # we don't want to update a side if only the opposing side's gene has been changed
      # https://dash.plotly.com/advanced-callbacks
      ctx = dash.callback_context

      common = self.callbackInputsToKwags(ctx.inputs_list[0])

      # the function used for creating the chart-figure is defined as a pointer in the views-object:
      fn = getattr(VIEWS, ctx.outputs_list['id']['view']).fn

      data = Utils.unstack_columns(fn(**common))
      
      # dash_pivottable.PivotTable looks nice, but doesnt work in our scenario very well...
      # https://dash.plotly.com/datatable
      return dash_table.DataTable(
           data=data.to_dict('records'),
          columns=[{"name": i, "id": i} for i in data.columns],
          fixed_rows={ 'headers': True, 'data': 0 },
          filter_action="native",
          sort_action="native",
          sort_mode="multi",
          style_cell={'textAlign': 'left'},
          style_cell_conditional=[
              {
                  'if': {'column_id': c},
                  'textAlign': 'right'
              } for c in data.columns if 'z-score' in c
          ],
          # virtualization=True, cant use it: https://community.plotly.com/t/dash-data-table-with-virtualization-seems-to-break-completely-when-hidden-on-startup/29328
          # make the datatable acquire all space of its container: https://community.plotly.com/t/dynamic-dash-datatable-height/44061
          # also see: assets/styles.css
          style_table={"height": "calc(100vh - 272px)",
           "width": "calc(100vw - 40px)",
           "overflow": "auto",
           "display": "flex",
           "flex-flow": "column"}, #{'overflow-x': 'auto'}, #calc(100vh - 120px)
          style_data_conditional=[
          {
              'if': {'row_index': 'odd'},
              'backgroundColor': 'rgb(248, 248, 248)'
          }
          ]
        )
      
  def heatmapByRegion(self, aggregation_function, gene, side):
    
    return heatmap(Comparison.byDonor(
      HumanMicroarrayData(gene).get(from_cache=True, aggregations=AGGREGATION_FUNCTIONS.data),
      MouseISHData(gene).get(from_cache=True, aggregations=AGGREGATION_FUNCTIONS.data),
      aggregation_function
    ))

  # TODO: enforce same scaling for both sides to make the visual data comparable
  def coexpressions(self, aggregation_function, structure_level, species, gene1, gene2, side):
    structure_level = f"level_{structure_level}"
    sizeBy = f"shared_{aggregation_function}"

    if(gene1 == gene2):
      raise Exception("Co-expressions for the same gene of the same species make no sense.")

    result1 = Comparison.species_map[species](gene1).get(from_cache=True, aggregations=AGGREGATION_FUNCTIONS.data)[species]

    result2 = Comparison.species_map[species](gene2).get(from_cache=True, aggregations=AGGREGATION_FUNCTIONS.data)[species]

    # sometimes, there is no coronal or sagittal data
        
    data = Comparison.coexpression(result1, result2, aggregation_function, structure_level, gene1, gene2) 

    # https://plotly.com/python-api-reference/generated/plotly.express.scatter
    fig = px.scatter(Utils.sort_case_insensitive(data, structure_level), 
          y=f"{Z_SCORE}_{gene1}_{aggregation_function}", 
          x=f"{Z_SCORE}_{gene2}_{aggregation_function}"
          # TODO: there is some issue with sizes. besides NaN, we got some weird errors...
	        #,size=sizeBy#, size_max=400 
          ,color=structure_level
          ,opacity=0.65
          # https://plotly.com/javascript/reference/#parcats-dimensions-items-dimension-categoryorder
          #,dimensions={'categoryorder':'category ascending'} # sort categories 
          )

    
    fig.update_yaxes(title=gene1)
    fig.update_xaxes(title=gene2)           

    return fig

  def stackedBarsBySpecies(self, aggregation_function, structure_level, species, gene, side):
    result = Comparison.species_map[species](gene).get(from_cache=True, aggregations=AGGREGATION_FUNCTIONS.data)[species]

    return heatmap(Comparison.by_region(
      result.structure, 
      aggregation_function, Z_SCORE, f'level_{structure_level}', 'structure_name')
    )

  def callbackInputsToKwags(self, input):
    return { input[i]['id']['type']:input[i]['value']  for i in range(0,len(input)) }

  def gridView(self, viewName, filters):
    return html.Div([
          dbc.Card(dbc.CardBody(
            dbc.Form( # we get an array (label + dropdown) inside an array. so we need to unpack that
              dbc.FormGroup(Utils.unpack(self.dropDownByList(lst, {'type': lst.type, 'view': viewName}, value=lst.default) for lst in filters)
              + self.downloadButton(id={'type': 'download-button', 'view': viewName},viewName=viewName))
            , inline=True), className="p-2")), 
            dbc.Row(
                dbc.Col(
            dcc.Loading(html.Div(id={ 'type': 'grid', 'view': viewName }, children=[], className="mt-3"
             ), color=self.loadingColor)
          ))], className="p-2", style={ 'width': 'calc(100vw - 24px)'}) 

  def getTableData(self, species, gene, structure_level):

    result = Comparison.species_map[species](gene).get(from_cache=True, aggregations=AGGREGATION_FUNCTIONS.data)[species]

    value_columns = [(Z_SCORE, agg) for agg in AGGREGATION_FUNCTIONS.data]
    agg_dict = { (Z_SCORE, agg): AGGREGATION_AGGREGATES[agg] for agg in AGGREGATION_FUNCTIONS.data }
    
    # structure_level is provided as a string, because it stems from a dropdown-callback... we need to cast it to int
    result = result.structure.groupby([f'level_{i}' for i in range(0, int(structure_level) + 1)], dropna=True)[value_columns].agg(agg_dict).dropna()
    
    result = Utils.drop_columns_if(result.reset_index(), ['INDEX'])
    
    self.downloads['grid'] = { 'filename': f"{species}_{gene}_{structure_level}", 'data': result }
    return result
  
  # TODO: two range-sliders (for x and y) on common-control-div to allow synchronization of axis. 
  # # TODO or maybe is it possible to callback to hidden controls to then trigger updates on the axis?
  def sideBySideView(self, viewName, commonFilters, sideFilters):
    dimensions = Utils.simple({ 'w': 'calc(49vw - 30px)', 'h': 'calc(100vh - 340px)' })

    return html.Div([
          dbc.Card(dbc.CardBody(
          dbc.Form( # we get an array (label + dropdown) inside an array. so we need to unpack that
            dbc.FormGroup(Utils.unpack(self.dropDownByList(lst, {'type': lst.type, 'view': viewName}, value=lst.default) for lst in commonFilters))
          , inline=True), className="p-2")),
          dbc.Row([
                dbc.Col([ # start of left col
                  dbc.Row(
                    dbc.Form(
                      dbc.FormGroup( # TODO: make this responsive, because they overlap on smaller screens
                        Utils.unpack(self.dropDownByList(lst, {'type': lst.type, 'side': 'left', 'view': viewName}, value=lst.defaultLeft) for lst in sideFilters)
                      ), inline=True, className="ml-4 mt-3")
                  ),
                  dcc.Loading([ # TODO: move loading-panel to also encapsulate alerts. then, they will disappear on change of parameters
                    dbc.Row(
                      dbc.Alert(children="", 
                      id={ 'type': 'alert', 'view': viewName, 'side': 'left'},
                      is_open=False, color="danger", className="ml-3 mr-2 mt-1 w-100"),
                    ),
                    dbc.Row(
                      html.Div(
                        dcc.Graph(id={ 'type': 'graph', 'view': viewName, 'side': 'left'}, config={'scrollZoom': True}, style={'width': dimensions.w, 'height': dimensions.h})
                      ,id={ 'type': 'graph-container', 'view': viewName, 'side': 'left'}), className="ml-2 mt-3")
                  ],color=self.loadingColor, className="w-100 h-100")
                ], className="border-right"), # end of left col
                dbc.Col([ # start of right col
                  dbc.Row(
                    dbc.Form(
                      dbc.FormGroup(
                        Utils.unpack(self.dropDownByList(lst, {'type': lst.type, 'side': 'right', 'view': viewName}, value=lst.defaultRight) for lst in sideFilters)
                      ), inline=True, className="ml-4 mt-3")
                  ),
                  dcc.Loading([
                  dbc.Row(
                    dbc.Alert(children="", 
                    id={ 'type': 'alert', 'view': viewName, 'side': 'right'},
                    is_open=False, color="danger", className="ml-2 mr-3 mt-1 w-100"),
                  ),
                  dbc.Row(html.Div(
                    dcc.Graph(id={ 'type': 'graph', 'view': viewName, 'side': 'right'}, config={'scrollZoom': True}, style={'width': dimensions.w, 'height': dimensions.h})
                 ,id={ 'type': 'graph-container', 'view': viewName, 'side': 'right'}), className="ml-2 mt-3")
                ] ,color=self.loadingColor)
                ]) # end of right col
          ])
        ], className="p-2", style={ 'width': 'calc(100vw - 8px)'})

  # https://community.plotly.com/t/gluphicons-in-bootstrap-buttons/23438/2
  def downloadButton(self, viewName, **kwags):
    return [dbc.Button([html.I(className="fas fa-download fa-lg")], **kwags), dcc.Loading(Download(id={'type': 'download', 'view': viewName}), color=self.loadingColor)]

  def dropDownByList(self, lst, idProps, **kwags):
    return ([dbc.Label(lst.label, className="mr-2")] if not lst.label is None else []) + [dbc.Select(id={'type': lst.type, 'input': True, **idProps}, 
      options=[{ "label": g, "value": g } for g in lst.data], className="mr-3", **kwags)]

  def run_server(self, **kwags):
    return self.app.run_server(**kwags, port=self.port)

  def __open(self):
    webbrowser.open_new("http://localhost:{}".format(self.port))

  def open_browser(self):
    Timer(1, self.__open).start()
    
# https://stackoverflow.com/questions/36898008/seaborn-heatmap-with-logarithmic-scale-colorbar
# => norm=LogNorm(), ...but this doesn't work with 0 or negative values
def heatmap(data, title = None, subplots_adjust_parameters = None, xlabel = None, ylabel = None, **kwags):

  fig = px.imshow(data)

  fig.update_xaxes(
        tickangle = 45)
  fig.update_layout(
    autosize=True,
    margin=dict(l=0, r=0, t=25, b=10),
    title=title)
  
  return fig




import os
import pandas as pd
import logging as log
import pickle
from types import SimpleNamespace
import itertools

class DropDownListConfiguration:
  def __init__(self, label, type, data, default, defaultLeft = None, defaultRight = None):
    self.label = label;
    self.type = type;
    self.data = data;
    self.default = default;
    self.defaultLeft = defaultLeft;
    self.defaultRight = defaultRight;
    
def z_score(data_col):
  # https://intellipaat.com/community/20492/pandas-compute-z-score-for-all-columns  
  # from https://www.statisticshowto.com/probability-and-statistics/z-score/: 
  # Simply put, a z-score (also called a standard score) gives you an idea of how far from the mean a data point is. 
  # But more technically it’s a measure of how many standard deviations below or above the population mean a raw score is.
  # https://community.brain-map.org/t/what-is-the-z-score-reference-value-in-rnaseq-gbm-data/513/3
  return (data_col - data_col.mean())/data_col.std()

def makedir(path):
  # from https://www.tutorialspoint.com/How-can-I-create-a-directory-if-it-does-not-exist-using-Python
  if not os.path.exists(path):
      os.makedirs(path)
  return path

def getFilename(filepath): 
  # https://stackoverflow.com/questions/4444923/get-filename-without-extension-in-python/4444952
  return os.path.splitext(os.path.basename(filepath))[0]

def getRelativeFilepath(path):
  return os.path.join(os.path.dirname(__file__), path)

def get_sub_columns(df, column):
  return [c[1] for c in df.columns if c[0]==column]

def drop_columns_if(df, keywords = ['structure_', 'level_']):
  # https://stackoverflow.com/questions/13411544/delete-column-from-pandas-dataframe
  ret = df.copy()

  for name in df.columns:
    if any(keyword in name[0] for keyword in keywords):
      ret = ret.drop(name, 1)
  
  return ret

def sort_by_nan(df):
  # from https://stackoverflow.com/questions/45909776/sort-rows-of-a-dataframe-in-descending-order-of-nan-counts
  return df.isnull().sum(1).sort_values(ascending=False).index

def sort_case_insensitive(df, column):
  # from: https://stackoverflow.com/questions/29898090/pandas-sort-with-capital-letters/29899345
  return df.loc[df[column].str.lower().sort_values().index]

def intersect(lst1, lst2): 
  # from https://www.geeksforgeeks.org/python-intersection-two-lists/
  return list(set(lst1) & set(lst2))
  
def merge_with_structure(data, structure, value_cols, aggregations):
  # merge while keeping each structure, even if there are no expression-levels.
  # https://stackoverflow.com/questions/31528819/using-merge-on-a-column-and-index-in-pandas
  ret = structure.merge(data,  left_index=True, right_on="structure_id", how="left")

  structure_identifier = ['structure_id', 'structure_name', 'acronym']
  level_cols = [col for col in ret.columns if 'level_' in col]
  byStructure = ret.groupby(level_cols + structure_identifier, dropna=False)[value_cols].agg(aggregations)
  byAcronym = ret.groupby('acronym', dropna=False)[value_cols].agg(aggregations)
  
  return simple({ 'structure': byStructure, 'acronym': byAcronym })

def combine_dicts(list_of_dicts):
  # https://stackoverflow.com/questions/3494906/how-do-i-merge-a-list-of-dicts-into-a-single-dict
  return {k: v for d in list_of_dicts for k, v in d.items()}

def splitByThreshold(data, column, separation_threshold):
  return (
    data[(data[column] < separation_threshold) & (data[column] > (-1 * separation_threshold))], 
    data[(data[column] > separation_threshold) | (data[column] < (-1 * separation_threshold))]);

def negativePart(number):
  return number if (number < 0) else 0

def save(obj, path, filename):
  # https://www.techcoil.com/blog/how-to-save-and-load-objects-to-and-from-file-in-python-via-facilities-from-the-pickle-module/
  makedir(path)
  with open(path + filename, 'wb') as file:
    # https://stackoverflow.com/questions/29127593/trying-to-write-a-cpickle-object-but-get-a-write-attribute-type-error
    pickle.dump(obj, file)

  return path + filename

def load(path):
  with open(path, 'rb') as file:
    return pickle.load(file)

def simple(dict):
  ret = SimpleNamespace()
  for k,v in dict.items():
    setattr(ret, k, v) 

  return ret
    
def dict_by_agg(data, aggregations, column_mappings):
  return { agg: data[[(experiment, agg) for experiment 
  in { **column_mappings['human'], **column_mappings['mouse']}.values()]].droplevel(1, axis=1) 
    for agg 
      in aggregations }    

def unstack_columns(df, separator = '_'):
  # inverse of: https://stackoverflow.com/questions/46247302/pandas-split-columns-into-multilevel
  return pd.DataFrame(df.values, columns=[separator.join([val for val in c if val != "" ]) for c in df.columns])  

def unpack(arg):
  # from https://stackoverflow.com/questions/13958998/python-list-comprehension-unpacking-and-multiple-operations
  # https://stackoverflow.com/questions/952914/how-to-make-a-flat-list-out-of-list-of-lists
  return [item for sublist in arg for item in sublist] #list(itertools.chain.from_iterable(arg))

# ? this helps profiling performance-bottlenecks
# from: https://towardsdatascience.com/how-to-profile-your-code-in-python-e70c834fad89
import cProfile
import pstats
from functools import wraps

def profile(output_file=None, sort_by='cumulative', lines_to_print=None, strip_dirs=False):
    """A time profiler decorator.
    Inspired by and modified the profile decorator of Giampaolo Rodola:
    http://code.activestate.com/recipes/577817-profile-decorator/
    Args:
        output_file: str or None. Default is None
            Path of the output file. If only name of the file is given, it's
            saved in the current directory.
            If it's None, the name of the decorated function is used.
        sort_by: str or SortKey enum or tuple/list of str/SortKey enum
            Sorting criteria for the Stats object.
            For a list of valid string and SortKey refer to:
            https://docs.python.org/3/library/profile.html#pstats.Stats.sort_stats
        lines_to_print: int or None
            Number of lines to print. Default (None) is for all the lines.
            This is useful in reducing the size of the printout, especially
            that sorting by 'cumulative', the time consuming operations
            are printed toward the top of the file.
        strip_dirs: bool
            Whether to remove the leading path info from file names.
            This is also useful in reducing the size of the printout
    Returns:
        Profile of the decorated function
    """

    def inner(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _output_file = output_file or func.__name__ + '.prof'
            pr = cProfile.Profile()
            pr.enable()
            retval = func(*args, **kwargs)
            pr.disable()
            pr.dump_stats(_output_file)

            with open(_output_file, 'w') as f:
                ps = pstats.Stats(pr, stream=f)
                if strip_dirs:
                    ps.strip_dirs()
                if isinstance(sort_by, (tuple, list)):
                    ps.sort_stats(*sort_by)
                else:
                    ps.sort_stats(sort_by)
                ps.print_stats(lines_to_print)
            return retval

        return wrapper

    return inner



from allensdk.api.queries.rma_api import RmaApi
from allensdk.api.queries.grid_data_api  import GridDataApi

from allensdk.api.queries.reference_space_api import ReferenceSpaceApi
from allensdk.core.reference_space import ReferenceSpace

import numpy as np
import pandas as pd
import glob

import concurrent
import concurrent.futures

from . import Utils
from . import Constants
from . import StructureMap

class MouseISHData:
  VALUE_COLUMNS = [Constants.EXPR_LVL, Constants.Z_SCORE] 

  currentGets = {}

  def __init__(self, geneAcronym):
    self.geneAcronym = geneAcronym
    self.cache_path = Constants.DATAFRAME_CACHE + f'mouse\\{geneAcronym}\\'

  def get(self, from_cache, aggregations):
  
    if from_cache:
      return self.getAsync(from_cache, aggregations)

    if self.geneAcronym in MouseISHData.currentGets:
      print(f'Waiting for initial request of mouse gene {self.geneAcronym} to complete...')
      done, not_done = concurrent.futures.wait([MouseISHData.currentGets[self.geneAcronym]], 
       return_when=concurrent.futures.FIRST_COMPLETED) # this wants an array... ok
      
      for fut in done:
        print(fut, 'exception:', fut.exception())
        return fut.result() 

    else: 
      with concurrent.futures.ThreadPoolExecutor() as executor:
        MouseISHData.currentGets[self.geneAcronym] = executor.submit(self.getAsync, from_cache, aggregations)
        return MouseISHData.currentGets[self.geneAcronym].result()

  def getAsync(self, from_cache, aggregations): 
    # load data once with from_cache = False, then change it to True to read it from disk instead of fetching it from the api
    if not from_cache:
      # we use the RmaApi to query specific information, such as the section data sets of a specific gene
      # for docs, see: https://alleninstitute.github.io/AllenSDK/allensdk.api.queries.rma_api.html
      rma = RmaApi() 
      
      # there might be a way to retrieve data in higher resolution, as stated here (default is 25, 10 is also available - but resolution is ignored for download_gene_expression_grid_data)
      # https://alleninstitute.github.io/AllenSDK/_modules/allensdk/api/queries/grid_data_api.html
      # See `Downloading 3-D Projection Grid Data <http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data#name="Downloading3-DExpressionGridData-DOWNLOADING3DPROJECTIONGRIDDATA">`_
      gdApi = GridDataApi()

      # http://api.brain-map.org/examples/rma_builder/index.html
      # http://api.brain-map.org/examples/rma_builder/rma_builder.html
      # https://allensdk.readthedocs.io/en/latest/data_api_client.html
      sectionDataSets = pd.DataFrame( 
          rma.model_query(                
                model='SectionDataSet',
                #! criteria="plane_of_section[name$eqcoronal]", note that saggital only spans the left hemisphere, so this is tough to compare with human data.
                filters={'failed':'false'},
                include=f"genes[acronym$il{self.geneAcronym}],products[id$eq1]", # $il = case-insensitive like | yes, weird notation... id = 1 = mouse brain atlas (not developing!)
                num_rows='all')
      )
       
      # model's documentation: http://api.brain-map.org/doc/SectionDataSet.html
      # https://community.brain-map.org/t/attempting-to-download-substructures-for-coronal-p56-mouse-atlas/174/2

      experiments = {}
      
      # http://help.brain-map.org/display/mousebrain/Documentation
      annotations = np.fromfile(Utils.getRelativeFilepath("annotations\\P56_Mouse_gridAnnotation\\gridAnnotation.raw"), dtype="uint32")

      # https://community.brain-map.org/t/how-to-acquire-the-structure-label-for-the-expression-grid-data/150/4
      # for Mouse P56, structure_graph_id = 1 according to http://help.brain-map.org/display/api/Atlas+Drawings+and+Ontologies
      structure_map = StructureMap.StructureMap(reference_space_key = 'annotation/ccf_2017', resolution=25).get(structure_graph_id=1) # , annotation, meta 
      # from http://alleninstitute.github.io/AllenSDK/_static/examples/nb/reference_space.html#Downloading-an-annotation-volume

      for index, row in sectionDataSets.iterrows(): # https://stackoverflow.com/questions/16476924/how-to-iterate-over-rows-in-a-dataframe-in-pandas
          exp_id = row['id']
          exp_path = f"cache\\mouse_ish-expr\\{exp_id}\\"
          
          try:
              # https://community.brain-map.org/t/whole-mouse-brain-gene-expression-data/447/4
              # explanation of what "energy" means here:
              # expression density = sum of expressing pixels / sum of all pixels in division
              # expression intensity = sum of expressing pixel intensity / sum of expressing pixels
              # expression energy = expression intensity * expression density

              gdApi.download_gene_expression_grid_data(exp_id, GridDataApi.ENERGY, exp_path)

              expression_levels = np.fromfile(exp_path + "energy.raw",  dtype=np.float32)

              # According to the doc @ http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data
              # we have "A raw uncompressed float (32-bit) little-endian volume representing average expression energy per voxel. 
              # A value of "-1" represents no data. This file is returned by default if the volumes parameter is null."
              data = pd.DataFrame({Constants.EXPR_LVL: expression_levels, "structure_id": annotations})

              # some expression_levels are assigned to a structure of id 0. same is true for Jure's approach.
              # according to the Allen institue, this is just due to background-noise: 
              # https://community.brain-map.org/t/how-to-acquire-the-structure-label-for-the-expression-grid-data/150/4
              # values of -1 mean "no value obtained", hence we filter them out:
              data = data[(data[Constants.EXPR_LVL] != -1) & (data.structure_id != 0)]

              data[Constants.Z_SCORE] = Utils.z_score(data[Constants.EXPR_LVL])

              # https://stackoverflow.com/questions/31528819/using-merge-on-a-column-and-index-in-pandas
              # https://stackoverflow.com/questions/45147100/pandas-drop-columns-with-all-nans                

              name = f'mouse_{exp_id}_{Constants.PlaneOfSections[row["plane_of_section_id"]]}'
              data = Utils.merge_with_structure(data, structure_map, MouseISHData.VALUE_COLUMNS, aggregations)

              Utils.save(data, self.cache_path, name + '.pkl')

              experiments['mouse - ' + Constants.PlaneOfSections[row["plane_of_section_id"]]] = data
          except Exception as e:
              print(f"Error retrieving mouse-ish experiment {exp_id}: {str(e)}")
              raise e

      return experiments
    else:
      if not glob.glob(self.cache_path):
        Utils.log.warning(f"No cached dataframe found. Check whether you have access to file '{self.cache_path}' and whether it exists. Obtaining data without caching now...")
        return self.get(False, aggregations)
      
      return { 'mouse - ' + Utils.getFilename(file).split('_')[2]: Utils.load(file) for file in glob.glob(f'{self.cache_path}/*.pkl') }          



from . import Visualisation

#def run():
webApp = Visualisation.WebInterface(__name__) 

webApp.run_server(debug=False)
#sys.exit()

#ärun()

# def run():
#   webApp = Visualisation.WebInterface(__name__) 

#   webApp.run_server(debug=False)
#   #sys.exit()


# https://www.pnas.org/content/107/28/12698.full
# https://genomebiology.biomedcentral.com/articles/10.1186/gb-2011-12-1-101
# https://sciencebasedmedicine.org/one-reason-mouse-studies-often-dont-translate-to-humans-very-well/
# https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3737272/
# https://blogs.sciencemag.org/pipeline/archives/2019/08/22/human-brains-and-mouse-brains-so-similar-so-different
# https://portal.brain-map.org/explore/transcriptome
# https://www.biorxiv.org/content/10.1101/384826v1.full
# https://viewer.cytosplore.org/ (only motor-cortex...)
# https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5055290/
# https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5299387/
# https://academic.oup.com/cercor/article/28/11/3829/4508772
# https://www.natureasia.com/en/research/highlight/13065#:~:text=Compared%20to%20the%20cerebral%20cortex,and%20the%20number%20of%20neurons.&text=Using%20mouse%20single%2Dcell%20RNA,corresponding%20cell%20types%20in%20mice.
# https://www.researchgate.net/figure/Anatomical-comparison-between-mouse-and-human-brains-The-comparison-between-rodent-A_fig1_221919008

# TODO: 
# match using manually created mapping of anatomical structures, e.g. based on Sequeira-paper and/or 
# https://science.sciencemag.org/content/367/6482/eaay5947/tab-figures-data (use link from google...)


# match using correlations of orthologues: https://www.proteinatlas.org/humanproteome/brain/mouse+brain

# compare euclidean distances between receptor-occurences per structure per species correlated against a (chosen) standard => see Sequeira fig 7, page 10 
# order regions in ateroposterior axis and use color-code as in fig 1 (sequeira)
# unsupervised clustering (e.g. ward's) of brain-regions on behalf of expression-levels

from functools import reduce
import pandas as pd
import numpy as np

from .HumanMicroarrayData import HumanMicroarrayData
from .MouseISHData import MouseISHData

from . import Constants
from . import Utils

species_map ={ 'human': HumanMicroarrayData, 'mouse': MouseISHData, 'mouse - coronal': MouseISHData, 'mouse - sagittal': MouseISHData} 

# merge a list of data-frames using a shared column. this column will be used as the new index. 
# uses an inner join. does not mutate the original data-frames.
def merge(dfs, dataset, merge_on, shared_columns):

  copies = []

  for el in dfs:
    for name, el_data in el.items(): # we only have one key-value pair, which is the name & data
      # we need to join using the index. otherwise, the column will be renamed during merge.
      data = getattr(el_data, dataset).reset_index().set_index(merge_on) # assigning this to a new var prevents mutating the original data-frames

      data = data[shared_columns]
      # drop na is fine. but we need to know when count = 0, so keep zeros
      data = data.dropna()

      # add_suffix is not suitable, as it would also affect sub-level columns (e.g. mean, var, etc. of expression-levels)
      # https://stackoverflow.com/questions/57740319/how-to-add-prefix-to-multi-index-columns-at-particular-level
      data = data.rename(mapper=lambda x: f'{x}_{name}', axis='columns', level=0)

      copies.append(data)
    
  return reduce(lambda  acc, next: pd.merge(acc, next, left_index=True, right_index=True, how='outer'), copies)


def union(dfs, keys =['human', 'mouse']):
  return pd.concat(dfs, keys=keys)

def by_region(structure_df, agg, value_column, region_column, structure_column):
  
  #aggregations = Utils.get_sub_columns(structure_df, value_column)
  
  region = structure_df.dropna()[[(value_column, agg)]].reset_index().droplevel(1, axis=1) 
  
  agg_fns = { 'min': np.min, 'max': np.max, 'mean': np.mean, 'var': np.mean } # we need to aggregate the columns differently...

  #for k, v in region_agg.items():
  region = region[[region_column, structure_column, value_column]].sort_values([region_column, value_column], ascending=True)
    # https://intellipaat.com/community/31330/pandas-number-rows-within-group
  region['rank'] = region.groupby([region_column]).cumcount() + 1
  region = pd.pivot_table(region, columns=['rank'], index=region_column, aggfunc=agg_fns[agg]).droplevel(0, axis=1)
    
    # pandas sorts case-sensitive, but we don't want this. so:
    # https://stackoverflow.com/questions/30521994/how-to-sort-row-index-case-insensitive-way-in-pandas-dataframe
  region = region.reindex(Utils.sort_by_nan(region)) # sorted(region.index, key=lambda x: x.lower()))
  
  return region

def merge_coex(struct1, struct2, genes, index_columns_to_keep = ['structure_name']):
  def simplify_index(s):
    return s.reset_index().set_index(index_columns_to_keep)[s.columns]

  return simplify_index(struct1).merge(simplify_index(struct2), left_index=True, right_index=True, suffixes=[ '_' + g for g in genes])

def findRegionAssignment(x, assignments, species):

  # ! its kind of a convention that levels_0 to 10 are the first 10 columns.
  # ! we need to iterate from highest to lowest level in order match the most specific region
  levels = [(f"level_{level}", x[level]) for level in range(10,0,-1)] 

  for l in levels:
    if l in assignments[species]:
      return assignments[species][l]['name']

  return None

def addRegionAssignments(df, species):
  df.reset_index(inplace=True)

  # ? comp contain joined data of human and mice. i.e. different species / experiments are provided as columns.
  # raw=True in order to only receive the ndarray instead of a Series. this is much faster, according to:
  # https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.apply.html#pandas.DataFrame.apply
  df[Constants.REGION_ASSIGNMENT] = df.apply(findRegionAssignment, axis=1, raw=True, args=(Constants.RegionAssignments.asDict, species, ))
  return df[df[Constants.REGION_ASSIGNMENT].notnull()]

def byDonor(human, mouse, agg, matchBy = Constants.REGION_ASSIGNMENT):
  
  human['human'].structure = addRegionAssignments(human['human'].structure, 'Human')
  mouse['mouse - sagittal'].structure = addRegionAssignments(mouse['mouse - sagittal'].structure, 'Mouse')
  mouse['mouse - coronal'].structure = addRegionAssignments(mouse['mouse - coronal'].structure, 'Mouse')

  comp = merge([human] + [mouse], 'structure', matchBy, Utils.intersect(HumanMicroarrayData.VALUE_COLUMNS, MouseISHData.VALUE_COLUMNS))

  # remove verbose structural details and remove glob-z-prefix to improve readability:
  comp = Utils.drop_columns_if(comp)

  # remove Z_SCORE-prefix to also improve readability:
  comp = comp.rename({ **column_mappings(mouse)['human'], **column_mappings(mouse)['mouse']},  axis='columns')

  # we isolate the aggregation
  return comp[[(experiment, agg) for experiment in { **column_mappings(mouse)['human'], **column_mappings(mouse)['mouse']}.values()]].droplevel(1, axis=1) 

def coexpression(data1, data2, aggregation_function, structure_level, gene1, gene2): # hemisphere, 
  
  data = merge_coex(data1.structure, data2.structure, [gene1, gene2], ['structure_name', structure_level]).reset_index().dropna()

  data[f'shared_{aggregation_function}'] = data[(f'{Constants.Z_SCORE}_{gene1}', aggregation_function)] * data[(f'{Constants.Z_SCORE}_{gene2}', aggregation_function)] 

  return Utils.unstack_columns(data)

def column_mappings(mouse_data):
  
  return { 
      'human': { Constants.Z_SCORE + '_' + 'human': 'human' }, 
      'mouse': { Constants.Z_SCORE + '_' + k:k.replace('_', ' ') for k,v in mouse_data.items() }
    } 


from . import Utils
from . import Constants
from . import StructureMap
from . import HumanMicroarrayData
from . import MouseISHData
from . import Comparison
from . import Visualisation
from . import App

